@@ -1,19 +1,19 @@
 from __future__ import unicode_literals
 
-from prompt_toolkit.filters import Condition, IsDone, HasCompletions, RendererHeightIsKnown
+from prompt_toolkit.enums import DEFAULT_BUFFER
+from prompt_toolkit.filters import IsDone, HasCompletions, RendererHeightIsKnown, Always, HasFocus
 from prompt_toolkit.key_binding.vi_state import InputMode
 from prompt_toolkit.layout import Window, HSplit, VSplit, FloatContainer, Float
 from prompt_toolkit.layout.controls import BufferControl, TokenListControl, FillControl
 from prompt_toolkit.layout.dimension import LayoutDimension
 from prompt_toolkit.layout.menus import CompletionsMenu
-from prompt_toolkit.layout.processors import BracketsMismatchProcessor
+from prompt_toolkit.layout.processors import HighlightSearchProcessor, HighlightSelectionProcessor, HighlightMatchingBracketProcessor, ConditionalProcessor
 from prompt_toolkit.layout.screen import Char
 from prompt_toolkit.layout.toolbars import CompletionsToolbar, ArgToolbar, SearchToolbar, ValidationToolbar, SystemToolbar, TokenListToolbar
 from prompt_toolkit.layout.utils import token_list_width
 from prompt_toolkit.selection import SelectionType
 
-from ptpython.filters import HasSignature, ShowCompletionsMenu, ShowCompletionsToolbar, ShowSidebar, ShowLineNumbersFilter, HadMultiplePythonBuffers, PythonBufferFocussed, ShowSignature, ShowDocstring
-from ptpython.utils import current_python_buffer
+from ptpython.filters import HasSignature, ShowCompletionsMenu, ShowCompletionsToolbar, ShowSidebar, ShowLineNumbersFilter, ShowSignature, ShowDocstring
 
 from pygments.lexers import PythonLexer
 from pygments.token import Token
@@ -46,21 +46,16 @@ class PythonSidebarControl(TokenListControl):
 
             def append(shortcut, label, status):
                 tokens.append((TB.Shortcut, ' [%s] ' % shortcut))
-                tokens.append((TB.Label, '%-18s' % label))
+                tokens.append((TB.Label, '%-21s' % label))
                 if status:
                     tokens.append((TB.Status, '%9s\n' % status))
                 else:
                     tokens.append((TB.Label, '\n'))
 
-            append('Ctrl-T', 'New tab', '')
-            append('Ctrl-D', 'Close tab', '')
-            append('Ctrl-Left/Right', 'Focus tab', '')
             append('F3', 'Completion menu', '(%s)' % completion_style)
             append('F4', 'Input mode', '(%s)' % mode)
-            append('F5', 'Show all tabs', '(on)' if settings.show_all_buffers else '(off)')
-#            append('F5', 'Merge from history', '')
+            append('F5', 'Complete while typing', '(on)' if settings.complete_while_typing else '(off)')
             append('F6', 'Paste mode', '(on)' if settings.paste_mode else '(off)')
-            append('F7', 'Multiline', '(always)' if settings.currently_multiline else '(auto)')
             append('F8', 'Show signature', '(on)' if settings.show_signature else '(off)')
             append('F9', 'Show docstring', '(on)' if settings.show_docstring else '(off)')
             append('F10', 'Show line numbers', '(on)' if settings.show_line_numbers else '(off)')
@@ -74,7 +69,7 @@ class PythonSidebar(Window):
     def __init__(self, settings, key_bindings_manager):
         super(PythonSidebar, self).__init__(
             PythonSidebarControl(settings, key_bindings_manager),
-            width=LayoutDimension.exact(34),
+            width=LayoutDimension.exact(37),
             filter=ShowSidebar(settings) & ~IsDone())
 
 
@@ -85,10 +80,8 @@ class SignatureControl(TokenListControl):
             append = result.append
             Signature = Token.Toolbar.Signature
 
-            _, python_buffer = current_python_buffer(cli, settings)
-
-            if python_buffer.signatures:
-                sig = python_buffer.signatures[0]  # Always take the first one.
+            if settings.signatures:
+                sig = settings.signatures[0]  # Always take the first one.
 
                 append((Signature, ' '))
                 try:
@@ -120,32 +113,6 @@ class SignatureControl(TokenListControl):
         super(SignatureControl, self).__init__(get_tokens)
 
 
-class TabsControl(TokenListControl):
-    """
-    Displays the list of tabs.
-    """
-    def __init__(self, settings):
-        def get_tokens(cli):
-            python_buffer_names = sorted([b for b in cli.buffers.keys() if b.startswith('python-')])
-
-            current_name, _ = current_python_buffer(cli, settings)
-
-            result = []
-            append = result.append
-
-            append((Token.TabBar, ' '))
-            for b in python_buffer_names:
-                if b == current_name:
-                    append((Token.TabBar.Tab.Active, ' %s ' % b))
-                else:
-                    append((Token.TabBar.Tab, ' %s ' % b))
-                append((Token.TabBar, ' '))
-
-            return result
-
-        super(TabsControl, self).__init__(get_tokens, Char(token=Token.TabBar), align_right=True)
-
-
 class SignatureToolbar(Window):
     def __init__(self, settings):
         super(SignatureToolbar, self).__init__(
@@ -162,14 +129,6 @@ class SignatureToolbar(Window):
                 ~IsDone())
 
 
-class TabsToolbar(Window):
-    def __init__(self, settings):
-        super(TabsToolbar, self).__init__(
-            TabsControl(settings),
-            height=LayoutDimension.exact(1),
-            filter=~IsDone() & HadMultiplePythonBuffers(settings))
-
-
 class PythonPrompt(TokenListControl):
     """
     Prompt showing something like "In [1]:".
@@ -184,9 +143,7 @@ class PythonPrompt(TokenListControl):
 class PythonToolbar(TokenListToolbar):
     def __init__(self, key_bindings_manager, settings, token=Token.Toolbar.Status):
         def get_tokens(cli):
-            _, python_buffer = current_python_buffer(cli, settings)
-            if not python_buffer:
-                return []
+            python_buffer = cli.buffers['default']
 
             TB = token
             result = []
@@ -214,7 +171,7 @@ class PythonToolbar(TokenListToolbar):
                 else:
                     append((TB.Off, '[F6] Paste mode (off)  '))
 
-                if python_buffer.is_multiline:
+                if python_buffer.is_multiline():
                     append((TB, ' [Meta+Enter] Execute'))
 
             return result
@@ -289,58 +246,50 @@ class ShowSidebarButtonInfo(Window):
             width=LayoutDimension.exact(width))
 
 
-def create_layout(buffers, settings, key_bindings_manager,
+def create_layout(settings, key_bindings_manager,
                   python_prompt_control=None, lexer=PythonLexer, extra_sidebars=None,
                   extra_buffer_processors=None):
     D = LayoutDimension
-    show_all_buffers = Condition(lambda cli: settings.show_all_buffers)
     extra_sidebars = extra_sidebars or []
     extra_buffer_processors = extra_buffer_processors or []
 
-    def create_buffer_window(buffer_name):
+    def create_python_input_window():
         def menu_position(cli):
             """
             When there is no autocompletion menu to be shown, and we have a signature,
             set the pop-up position at `bracket_start`.
             """
-            b = cli.buffers[buffer_name]
+            b = cli.buffers['default']
 
-            if b.complete_state is None and b.signatures:
-                row, col =  b.signatures[0].bracket_start
+            if b.complete_state is None and settings.signatures:
+                row, col = settings.signatures[0].bracket_start
                 index = b.document.translate_row_col_to_index(row - 1, col)
                 return index
 
         return Window(
             BufferControl(
-                buffer_name=buffer_name,
+                buffer_name=DEFAULT_BUFFER,
                 lexer=lexer,
-                show_line_numbers=ShowLineNumbersFilter(settings, buffer_name),
-                input_processors=[BracketsMismatchProcessor()] + extra_buffer_processors,
+                show_line_numbers=ShowLineNumbersFilter(settings, 'default'),
+                input_processors=[
+                                  # Show matching parentheses, but only while editing.
+                                  ConditionalProcessor(
+                                      processor=HighlightMatchingBracketProcessor(chars='[](){}'),
+                                      filter=HasFocus(DEFAULT_BUFFER) & ~IsDone()),
+                                  HighlightSearchProcessor(preview_search=Always()),
+                                  HighlightSelectionProcessor()] + extra_buffer_processors,
                 menu_position=menu_position,
+
+                # Make sure that we always see the result of an reverse-i-search:
+                preview_search=Always(),
             ),
-            # As long as we're editing, prefer a minimal height of 8.
+            # As long as we're editing, prefer a minimal height of 6.
             get_height=(lambda cli: (None if cli.is_done else D(min=6))),
-
-            # When done, show only if this was focussed.
-            filter=(~IsDone() & show_all_buffers) | PythonBufferFocussed(buffer_name, settings)
         )
 
-    def create_buffer_window_separator(buffer_name):
-        return Window(
-            width=D.exact(1),
-            content=FillControl('\u2502', token=Token.Separator),
-            filter=~IsDone() & show_all_buffers)
-
-    buffer_windows = []
-    for b in sorted(buffers):
-        if b.startswith('python-'):
-            buffer_windows.append(create_buffer_window_separator(b))
-            buffer_windows.append(create_buffer_window(b))
-
     return HSplit([
         VSplit([
             HSplit([
-                TabsToolbar(settings),
                 FloatContainer(
                     content=HSplit([
                         VSplit([
@@ -348,7 +297,7 @@ def create_layout(buffers, settings, key_bindings_manager,
                                 python_prompt_control,
                                 dont_extend_width=True,
                             ),
-                            VSplit(buffer_windows),
+                            create_python_input_window(),
                         ]),
                     ]),
                     floats=[
